/*
 *  GIFTranslator.cpp 1.1.0 (Jan 3rd, 1998)
 *
 *  GIF translator add-on
 *  Written by Edmund Vermeulen <edmundv@xs4all.nl>
 *  Public domain
 */


/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ByteOrder.h>
#include <TypeConstants.h>
#include <DataIO.h>
#include <TranslatorAddOn.h>
#include <TranslatorFormats.h>
#include <InterfaceDefs.h>


/* Defines */
#define GIF_FORMAT 'GIF '


/* Handler information */
char translatorName[] = "GIFTranslator";
char translatorInfo[] = "By Edmund Vermeulen <edmundv@xs4all.nl>";
int32 translatorVersion = 110;

/* Supported input formats */
translation_format inputFormats[] =
{
	{ GIF_FORMAT, B_TRANSLATOR_BITMAP, 0.3, 0.5, "image/gif", "GIF" },
	{ 0, 0, 0.0, 0.0, "", "" }
};

/* Supported output formats */
translation_format outputFormats[] =
{
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.3, 0.5, "image/x-be-bitmap", "Be bitmap" },
	{ 0, 0,	0.0, 0.0, "", "" }
};


/* GIF Stuff */
#define MAXCOLORMAPSIZE 256

#define TRUE     1
#define FALSE    0

#define CM_RED   0
#define CM_GREEN 1
#define CM_BLUE  2

#define MAX_LWZ_BITS 12

#define INTERLACE     0x40
#define LOCALCOLORMAP 0x80
#define BitSet(byte, bit)    (((byte) & (bit)) == (bit))

#define ReadOK(buffer, len) (inStream->Read(buffer, len) == len)

#define LM_to_uint(a, b)     (((b)<<8)|(a))

const int N_TAGS = 1;

class GifDecoder
{
public:
	status_t ReadGIF(BPositionIO *in_stream, BPositionIO *out_stream, uint32 outFormat);

private:
	int ReadColorMap(int number, unsigned char buffer[3][MAXCOLORMAPSIZE]);
	int DoExtension(int label);
	int GetDataBlock(unsigned char *buf);
	int GetCode(int code_size, int flag);
	int LWZReadByte(int flag, int input_code_size);

	BPositionIO *inStream, *outStream;
	uint32 fOutFormat;
	int ZeroDataBlock;
	char ext_buf[256];
	unsigned char get_buf[280];
	int curbit, lastbit, done, last_byte;
	int fresh;
	int code_size, set_code_size;
	int max_code, max_code_size;
	int firstcode, oldcode;
	int clear_code, end_code;
	int table[2][(1<< MAX_LWZ_BITS)];
	int stack[(1<<(MAX_LWZ_BITS))*2], *sp;
};

status_t
GifDecoder::ReadGIF(BPositionIO *in_stream, BPositionIO *out_stream, uint32 outFormat)
{
	status_t        ret_val = B_OK;
	int             imageNumber = 1;
	unsigned char   buf[16];
	unsigned char   c;
	unsigned char   localColorMap[3][MAXCOLORMAPSIZE];
/*	int             grayScale;*/
	int             useGlobalColormap;
	int             bitPixel;
	int             imageCount = 0;
	char            version[4];
	struct {
		unsigned int    Width;
		unsigned int    Height;
		unsigned char   ColorMap[3][MAXCOLORMAPSIZE];
		unsigned int    BitPixel;
		unsigned int    ColorResolution;
		unsigned int    Background;
		unsigned int    AspectRatio;
		/*
		**
		*/
		int             GrayScale;
	} GifScreen;
	unsigned char (*cmap)[MAXCOLORMAPSIZE];
	uchar *row_buffer = NULL;

	last_byte = 2;  // EV: avoid using an unintialised variable

	inStream = in_stream;
	outStream = out_stream;
	fOutFormat = outFormat;

	ZeroDataBlock = FALSE;
	fresh = FALSE;

	if (! ReadOK(buf, 6))
		puts("error reading magic number" );

	if (strncmp((char *)buf, "GIF", 3) != 0)
		puts("not a GIF file" );

	strncpy(version, (char *)buf + 3, 3);
	version[3] = '\0';

	if ((strcmp(version, "87a") != 0) && (strcmp(version, "89a") != 0))
		puts("bad version number, not '87a' or '89a'" );

	if (! ReadOK(buf, 7))
		puts("failed to read screen descriptor" );

	GifScreen.Width           = LM_to_uint(buf[0], buf[1]);
	GifScreen.Height          = LM_to_uint(buf[2], buf[3]);
	GifScreen.BitPixel        = 2<<(buf[4]&0x07);
	GifScreen.ColorResolution = (((buf[4]&0x70)>>3)+1);
	GifScreen.Background      = buf[5];
	GifScreen.AspectRatio     = buf[6];

	/* Fill in the header and write it */
	TranslatorBitmap bmp;
	bmp.magic = B_TRANSLATOR_BITMAP;
	bmp.bounds.left = 0.0;
	bmp.bounds.top = 0.0;
	bmp.bounds.right = (float) GifScreen.Width - 1.0;
	bmp.bounds.bottom = (float) GifScreen.Height - 1.0;
	bmp.colors = B_RGB_32_BIT;
	bmp.rowBytes = GifScreen.Width * 4;
	bmp.dataSize = bmp.rowBytes * GifScreen.Height;

	/* First convert the header to the correct endianess */
	swap_data(B_UINT32_TYPE, &(bmp.magic), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_RECT_TYPE, &(bmp.bounds), sizeof(BRect), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_UINT32_TYPE, &(bmp.rowBytes), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_UINT32_TYPE, &(bmp.colors), sizeof(color_space), B_SWAP_HOST_TO_BENDIAN);
	swap_data(B_UINT32_TYPE, &(bmp.dataSize), sizeof(uint32), B_SWAP_HOST_TO_BENDIAN);

	outStream->Write(&bmp, sizeof(TranslatorBitmap));

	if (BitSet(buf[4], LOCALCOLORMAP)) {    /* Global Colormap */
		if (ReadColorMap(GifScreen.BitPixel, GifScreen.ColorMap))
			puts("error reading global colormap" );
	}

	if (GifScreen.AspectRatio != 0 && GifScreen.AspectRatio != 49) {
		float   r;
		r = ( (float) GifScreen.AspectRatio + 15.0 ) / 64.0;
		printf("warning - non-square pixels; to fix do a 'pnmscale -%cscale %g'\n",
			r < 1.0 ? 'x' : 'y',
			r < 1.0 ? 1.0 / r : r );
	}

	for (;;) {
		if (! ReadOK(&c, 1))
			puts("EOF / read error on image data" );

		if (c == ';') {         /* GIF terminator */
			if (imageCount < imageNumber)
				printf("only %d image%s found in file\n", imageCount, imageCount>1?"s":"" );
			return ret_val;
		}

		if (c == '!') {         /* Extension */
			if (! ReadOK(&c, 1))
				puts("OF / read error on extention function code");
			DoExtension(c);
			continue;
		}

		if (c != ',') {         /* Not a valid start character */
			printf("bogus character 0x%02x, ignoring\n", (int) c );
			continue;
		}

		++imageCount;

		if (! ReadOK(buf, 9))
			puts("couldn't read left/top/width/height");

		useGlobalColormap = ! BitSet(buf[8], LOCALCOLORMAP);

		bitPixel = 1<<((buf[8]&0x07)+1);

		if (! useGlobalColormap) {
			if (ReadColorMap(bitPixel, localColorMap))
				puts("error reading local colormap" );
			cmap = localColorMap;
		} else {
			cmap = GifScreen.ColorMap;
		}

		int left = LM_to_uint(buf[0], buf[1]);
		int top = LM_to_uint(buf[2], buf[3]);
		int width = LM_to_uint(buf[4], buf[5]);
		int height = LM_to_uint(buf[6], buf[7]);
		int interlace = BitSet(buf[8], INTERLACE);
		int ignore = (imageCount != imageNumber) && (fOutFormat == B_TRANSLATOR_BITMAP);

		int v;
		int xpos = left, ypos = top, pass = 0;

		/*
		**  Initialize the Compression routines
		*/
		if (! ReadOK(&c, 1))
			puts("EOF / read error on image data" );

		if (LWZReadByte(TRUE, c) < 0)
			puts("error reading image" );

		/*
		**  If this is an "uninteresting picture" ignore it.
		*/
		if (ignore) {
			puts("skipping image..." );
			while (LWZReadByte(FALSE, c) >= 0)
				;
		}
		else
		{
			printf("reading %d by %d%s GIF image, offset %d, %d\n", width, height, interlace ? " interlaced" : "", left, top );
	
			long buf_size = width * 4;
			if (interlace)
				buf_size *= 8;
			if ((row_buffer = (uchar *) malloc(buf_size)) == NULL) {
				puts("couldn't alloc space for image" );
				return B_ERROR;
			}
			memset(row_buffer, 0, buf_size);

			if (!interlace)
				pass = 3;
	
			uchar *in_row = row_buffer;
			while ((v = LWZReadByte(FALSE, c)) >= 0 ) {
				in_row[0] = cmap[CM_BLUE][v];
				in_row[1] = cmap[CM_GREEN][v];
				in_row[2] = cmap[CM_RED][v];
				in_row[3] = 255;  // Set alpha to fully opaque
				in_row += 4;
	
				++xpos;
				if (xpos == left + width) {
					if (interlace)
						outStream->Seek(sizeof(TranslatorBitmap) + ypos * 4 * width, SEEK_SET);
					if (outStream->Write(row_buffer, width * 4) < B_OK)
						ret_val = B_ERROR;

					in_row = row_buffer;
					xpos = left;
					if (interlace) {
						switch (pass) {
						case 0:
						case 1:
							ypos += 8; break;
						case 2:
							ypos += 4; break;
						case 3:
							ypos += 2; break;
						}
	
						if (ypos >= top + height) {
							++pass;
							switch (pass) {
							case 1:
								ypos = top + 4; break;
							case 2:
								ypos = top + 2; break;
							case 3:
								ypos = top + 1; break;
							default:
								goto fini;
							}
						}
					} else {
						++ypos;
					}
				}
				if ((ypos == top + height) || (ret_val != B_OK))
					break;
			}
			if (ret_val != B_OK)
				break;

fini:
			if (LWZReadByte(FALSE, c) >= 0)
				puts("too much input data, ignoring extra...");
			free(row_buffer);
		}
	}
	return ret_val;
}

int
GifDecoder::ReadColorMap(int number, unsigned char buffer[3][MAXCOLORMAPSIZE])
{
	int             i;
	unsigned char   rgb[3];

	for (i = 0; i < number; ++i) {
		if (! ReadOK(rgb, sizeof(rgb)))
			puts("bad colormap" );

		buffer[CM_RED][i]   = rgb[0];
		buffer[CM_GREEN][i] = rgb[1];
		buffer[CM_BLUE][i]  = rgb[2];
	}
	return FALSE;
}

int
GifDecoder::DoExtension(int label)
{
	char            *str;
/*
	struct {
		int     transparent;
		int     delayTime;
		int     inputFlag;
		int     disposal;
	} Gif89 = { -1, -1, -1, 0 };
*/

	switch (label) {
	case 0x01:              /* Plain Text Extension */
		str = "Plain Text Extension";
		break;
	case 0xff:              /* Application Extension */
		str = "Application Extension";
		break;
	case 0xfe:              /* Comment Extension */
		str = "Comment Extension";
		while (GetDataBlock((unsigned char*) ext_buf) != 0) {
			printf("gif comment: %s\n", ext_buf );
		}
		return FALSE;
	case 0xf9:              /* Graphic Control Extension */
		str = "Graphic Control Extension";
		(void) GetDataBlock((unsigned char*) ext_buf);
/*
		Gif89.disposal    = (ext_buf[0] >> 2) & 0x7;
		Gif89.inputFlag   = (ext_buf[0] >> 1) & 0x1;
		Gif89.delayTime   = LM_to_uint(ext_buf[1], ext_buf[2]);
		if ((ext_buf[0] & 0x1) != 0)
		Gif89.transparent = ext_buf[3];
*/
		while (GetDataBlock((unsigned char*) ext_buf) != 0)
			;
		return FALSE;
	default:
		str = ext_buf;
		sprintf(ext_buf, "UNKNOWN (0x%02x)\n", label);
		break;
	}

	printf("got a '%s' extension\n", str );

	while (GetDataBlock((unsigned char*) ext_buf) != 0)
		;

	return FALSE;
}

int
GifDecoder::GetDataBlock(unsigned char *buf)
{
	unsigned char   count;

	if (! ReadOK(&count, 1)) {
		puts("error in getting DataBlock size" );
		return -1;
	}

	ZeroDataBlock = (count == 0);

	if ((count != 0) && (! ReadOK(buf, count))) {
		puts("error in reading DataBlock" );
	return -1;
	}

	return count;
}

int
GifDecoder::GetCode(int code_size, int flag)
{
	int             i, j, ret;
	unsigned char   count;

	if (flag) {
		curbit = 0;
		lastbit = 0;
		done = FALSE;
		return 0;
	}

	if ( (curbit+code_size) >= lastbit) {
		if (done) {
			if (curbit >= lastbit)
				puts("ran off the end of my bits" );
			return -1;
		}
		get_buf[0] = get_buf[last_byte-2];
		get_buf[1] = get_buf[last_byte-1];

		if ((count = GetDataBlock(&get_buf[2])) == 0)
			done = TRUE;

		last_byte = 2 + count;
		curbit = (curbit - lastbit) + 16;
		lastbit = (2+count)*8 ;
	}

	ret = 0;
	for (i = curbit, j = 0; j < code_size; ++i, ++j)
		ret |= ((get_buf[ i / 8 ] & (1 << (i % 8))) != 0) << j;

	curbit += code_size;

	return ret;
}

int
GifDecoder::LWZReadByte(int flag, int input_code_size)
{
	int             code, incode;
	register int    i;

	if (flag) {
		set_code_size = input_code_size;
		code_size = set_code_size+1;
		clear_code = 1 << set_code_size ;
		end_code = clear_code + 1;
		max_code_size = 2*clear_code;
		max_code = clear_code+2;

		GetCode(0, TRUE);
		fresh = TRUE;

		for (i = 0; i < clear_code; ++i) {
			table[0][i] = 0;
			table[1][i] = i;
		}
		for (; i < (1<<MAX_LWZ_BITS); ++i)
			table[0][i] = table[1][0] = 0;

		sp = stack;

		return 0;
	} else if (fresh) {
		fresh = FALSE;
		do {
			firstcode = oldcode = GetCode(code_size, FALSE);
		} while (firstcode == clear_code);
		return firstcode;
	}

	if (sp > stack)
		return *--sp;

	while ((code = GetCode(code_size, FALSE)) >= 0) {
		if (code == clear_code) {
			for (i = 0; i < clear_code; ++i) {
				table[0][i] = 0;
				table[1][i] = i;
			}
			for (; i < (1<<MAX_LWZ_BITS); ++i)
				table[0][i] = table[1][i] = 0;
			code_size = set_code_size+1;
			max_code_size = 2*clear_code;
			max_code = clear_code+2;
			sp = stack;
			firstcode = oldcode = GetCode(code_size, FALSE);
			return firstcode;
		} else if (code == end_code) {
			int             count;
			unsigned char   buf[260];

			if (ZeroDataBlock)
				return -2;

			while ((count = GetDataBlock(buf)) > 0)
				;

			if (count != 0)
				puts("missing EOD in data stream (common occurence)");
			return -2;
		}

		incode = code;

		if (code >= max_code) {
			*sp++ = firstcode;
			code = oldcode;
		}

		while (code >= clear_code) {
			*sp++ = table[1][code];
			if (code == table[0][code])
				puts("circular table entry BIG ERROR");
			code = table[0][code];
		}

		*sp++ = firstcode = table[1][code];

		if ((code = max_code) <(1<<MAX_LWZ_BITS)) {
			table[0][code] = oldcode;
			table[1][code] = firstcode;
			++max_code;
			if ((max_code >= max_code_size) &&
			    (max_code_size < (1<<MAX_LWZ_BITS))) {
				max_code_size *= 2;
				++code_size;
			}
		}

		oldcode = incode;

		if (sp > stack)
			return *--sp;
	}
	return code;
}


/* Identify function */
status_t
Identify(
	BPositionIO *inSource,
	const translation_format *inFormat,
	BMessage * /*ioExtension*/,
	translator_info *outInfo,
	uint32 outType)
{
	if (outType && outType != B_TRANSLATOR_BITMAP)
		return B_NO_TRANSLATOR;

	/* Read in the header */
	char data[6];
	if (inSource->Read(data, 6) != 6)
		return B_NO_TRANSLATOR;

	/* Check if it is GIF */
	if (strncmp(data, "GIF87a", 6) != 0 && strncmp(data, "GIF89a", 6) != 0)
		return B_NO_TRANSLATOR;

	inFormat = &inputFormats[0];

	/* Set the translator_info to the recognized type */
	outInfo->type = inFormat->type;
	outInfo->group = inFormat->group;
	outInfo->quality = inFormat->quality;
	outInfo->capability = inFormat->capability;
	strcpy(outInfo->name, inFormat->name);
	strcpy(outInfo->MIME, inFormat->MIME);
	return B_OK;
}


/* Translate function */
status_t
Translate(
	BPositionIO *inSource,
	const translator_info *inInfo,
	BMessage * /*ioExtension*/,
	uint32 outFormat,
	BPositionIO *outDestination)
{
	status_t ret_val = B_NO_TRANSLATOR;

	/* Default output format */
	if (outFormat == 0)
		outFormat = B_TRANSLATOR_BITMAP;

	/* Read a GIF image */
	if (inInfo->type == GIF_FORMAT && outFormat == B_TRANSLATOR_BITMAP)
	{
		GifDecoder *my_decoder = new GifDecoder;
		ret_val = my_decoder->ReadGIF(inSource, outDestination, outFormat);
		delete my_decoder;
	}

	return ret_val;
}
